/*
** Module   : MsSqlHelper.java
** Abstract : Helper class which initializes MSSQL 2012 database
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 OM  20140307 Created initial version. Helper class which provides utilities for
**                  Microsoft SQL server 2012.
** 002 OM  20140728 Added support for HQLFunction annotated methods. Granted exec acess to 
**                  defined UDFs. Switched nvarchar(MAX) for String parameters and double 
**                  precision for floating point numerical values.
*/

package com.goldencode.p2j.persist.dialect;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.pl.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Helper class which provides MSSQL-specific services: TBA
 */
public final class MsSqlHelper
{
   /** The name of the assembly that maps SQL functions and datatypes to CLR. */
   private static final String P2J2CLR = "p2j2clr";
   
   /**
    * Private constructor. All access to this class is via static methods.
    */
   private MsSqlHelper()
   {
   }
   
   /**
    * Register the function aliases with the HQL preprocessor, for the permanent database.
    *
    * @param   database
    *          Database to prepare.
    */
   public static void preparePermanentDatabase(Database database)
   {
      getFunctionAliases(database, BuiltIns.getPublicFunctions(), true, true);
      getFunctionAliases(database, BuiltIns.getSupportFunctions(), false, true);
   }
   
   /**
    * Get a list of database prepare DDL statements.
    *
    * @param   database
    *          Database to prepare.
    *
    * @return  the list of DDL statements.
    */
   public static List<String> getPrepareStatements(Database database)
   {
      // Create built-in function aliases.
      return getFunctionAliases(database, BuiltIns.getPublicFunctions(), true, false);
   }
   
   /**
    * Get function aliases for the given list of static methods, optionally making the alias names
    * unique by appending parameter type decorations.
    *
    * @param   database
    *          Database in which to create aliases.
    * @param   funcs
    *          List of static methods for which function aliases should be created.
    * @param   overload
    *          <code>true</code> to create unique function names for each overloaded method.
    * @param   register
    *          <code>true</code> if the function should be registered with
    *          the {@link com.goldencode.p2j.persist.HQLPreprocessor}.
    *
    * @return  the list of function aliases.
    */
   private static List<String> getFunctionAliases(Database database,
                                                  List<Method> funcs,
                                                  boolean overload,
                                                  boolean register)
   {
      List<String> ddl = new ArrayList<>();
      
      for (Method method : funcs)
      {
         HQLFunction hqlFn = method.getAnnotation(HQLFunction.class);
         if (hqlFn == null)
         {
            // not a support function
            continue;
         }
         
         String sqlUniqueName = method.getName();
         String hqlName = hqlFn.name().isEmpty() ? method.getName() : hqlFn.name();
         StringBuilder functionDecoration = new StringBuilder("_");
         StringBuilder paramList = new StringBuilder();
         
         Class<?>[] signature = method.getParameterTypes();
         for (int i = 0; i < signature.length; i++)
         {
            if (i > 0)
            {
               paramList.append(", ");
            }
            String paramType = signature[i].getName();
            paramList.append("@param").append(i + 1);
            paramList.append(" ").append(getSqlType(paramType));
            functionDecoration.append(getSqlDecoration(paramType));
         }
         
         if (overload)
         {
            sqlUniqueName += functionDecoration.toString();
         }
         
         ddl.add("\ncreate function [dbo].[" + sqlUniqueName + "]" +
                 "\n    (" + paramList.toString() + ")" +
                 "\n    returns " + getSqlType(method.getReturnType().getName()) +
                 "\n    as external name [" + P2J2CLR + "].[" +
                       method.getDeclaringClass().getSimpleName() + "].[" + sqlUniqueName + "]" +
                 "\ngo" +           // only one CREATE FUNCTION in a batch is allowed
                 "\ngrant exec on [dbo].[" + sqlUniqueName + "] to public" +
                 "\ngo");            // allow the function to be executed by all database users
         
         if (overload && register)
         {
            // drop the [ / ] in order to pass through Hibernate
            HQLPreprocessor.registerFunction(database, hqlName,  "dbo." + sqlUniqueName, method);
         }
      }
      
      return ddl;
   }

   /**
    * Returns the MS SQL Server datatype associated with the java type.
    *
    * @param   javaType
    *          The java type (full name).
    *
    * @return  The MS SQL Server datatype mapped by <code>javaType</code>.
    *          If <code>javaType</code> is invalid, <code>null</code> is returned.
    */
   private static String getSqlType(String javaType)
   {
      switch (javaType)
      {
         case "java.lang.Boolean":
         case "boolean":
            return "bit";
            
         case "java.lang.Integer":
            return "int";
            
         case "java.lang.Long":
            return "bigint";
            
         case "java.lang.String":
            return "nvarchar(MAX)";    // CLR string maps to nvarchar !
            
         case "java.math.BigDecimal":
            // return "numeric(38, 10)";  // tough decision regarding the precision
            return "double precision";
            
         case "java.sql.Date":
            return "date";
            
         case "java.sql.Timestamp":
            return "datetime2";
      }
      return null;
   }

   /**
    * Returns the parameter decoration for fixing the function overloading constraint in
    * MS SQL Server.
    *
    * @param   javaType
    *          The java type (full name).
    *
    * @return  a character for decorating the SQL function.
    */
   private static char getSqlDecoration(String javaType)
   {
      switch (javaType)
      {
         case "java.lang.Boolean":
         case "boolean":
            return 'b';
            
         case "java.lang.Integer":
            return 'i';
            
         case "java.lang.Long":
            return 'l';
            
         case "java.lang.String":
            return 's';
            
         case "java.math.BigDecimal":
            return 'r';
            
         case "java.sql.Date":
            return 'd';
            
         case "java.sql.Timestamp":
            return 't';
      }
      return ' ';
   }
}
